# =============================================================================
# Minet Fetch CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and fetching the urls found
# in the given column. This is done in a respectful multithreaded fashion to
# optimize both running time & memory.
#
import casanova
from io import StringIO
from collections import Counter
from ural import is_shortened_url
from ebbe.decorators import with_defer

from minet.fetch import multithreaded_fetch, multithreaded_resolve
from minet.fs import FilenameBuilder, ThreadSafeFilesWriter
from minet.web import grab_cookies, parse_http_header
from minet.exceptions import InvalidURLError, FilenameFormattingError
from minet.cli.exceptions import InvalidArgumentsError
from minet.cli.reporters import report_error, report_filename_formatting_error
from minet.cli.utils import LoadingBar, die


FETCH_ADDITIONAL_HEADERS = [
    "resolved",
    "status",
    "error",
    "filename",
    "mimetype",
    "encoding",
]

RESOLVE_ADDITIONAL_HEADERS = ["resolved", "status", "error", "redirects", "chain"]


@with_defer()
def fetch_action(cli_args, resolve=False, defer=None):

    # If we are hitting a single url we enable contents_in_report by default
    if (
        not resolve
        and isinstance(cli_args.file, StringIO)
        and cli_args.contents_in_report is None
    ):
        cli_args.contents_in_report = True

    if not resolve and cli_args.contents_in_report and cli_args.compress:
        raise InvalidArgumentsError(
            "Cannot both --compress and output --contents-in-report!"
        )

    # HTTP method
    http_method = cli_args.method

    # Cookie grabber
    get_cookie = None
    if cli_args.grab_cookies:
        get_cookie = grab_cookies(cli_args.grab_cookies)

    # Global headers
    global_headers = None
    if cli_args.headers:
        global_headers = {}

        for header in cli_args.headers:
            k, v = parse_http_header(header)
            global_headers[k] = v

    # Resume listener
    skipped_rows = 0
    resuming_reader_loading = None

    if cli_args.resume and cli_args.output.can_resume():
        resuming_reader_loading = LoadingBar(desc="Resuming", unit="line")

        def output_read_listener(event, row):
            nonlocal skipped_rows

            if event != "output.row":
                return

            skipped_rows += 1
            resuming_reader_loading.update()

        cli_args.output.listener = output_read_listener

    if resolve:
        additional_headers = RESOLVE_ADDITIONAL_HEADERS
    else:
        additional_headers = FETCH_ADDITIONAL_HEADERS

        if cli_args.contents_in_report:
            additional_headers = additional_headers + ["raw_contents"]

    # Enricher
    multiplex = None

    if cli_args.separator is not None:
        multiplex = (cli_args.column, cli_args.separator)

    enricher = casanova.threadsafe_enricher(
        cli_args.file,
        cli_args.output,
        add=additional_headers,
        keep=cli_args.select,
        total=cli_args.total,
        multiplex=multiplex,
    )

    if resuming_reader_loading is not None:
        resuming_reader_loading.close()

    if cli_args.column not in enricher.headers:
        raise InvalidArgumentsError(
            'Could not find the "%s" column containing the urls in the given CSV file.'
            % cli_args.column
        )

    url_pos = enricher.headers[cli_args.column]

    filename_pos = None

    if not resolve and cli_args.filename is not None:
        if cli_args.filename not in enricher.headers:
            raise InvalidArgumentsError(
                'Could not find the "%s" column containing the filenames in the given CSV file.'
                % cli_args.filename
            )

        filename_pos = enricher.headers[cli_args.filename]

    # Loading bar
    loading_bar = LoadingBar(
        desc="Fetching pages", total=enricher.total, unit="url", initial=skipped_rows
    )
    defer(
        loading_bar.close
    )  # NOTE: it could be dangerous with multithreaded execution, not to close it ourselves

    def update_loading_bar(result):
        nonlocal errors

        if result.error is not None:
            errors += 1
        else:
            if resolve:
                status = result.stack[-1].status
            else:
                status = result.response.status
            if status is not None and status >= 400:
                status_codes[status] += 1

        stats = {"errors": errors}

        for code, count in status_codes.most_common(1):
            stats[str(code)] = count

        loading_bar.update_stats(**stats)
        loading_bar.update()

    only_shortened = getattr(cli_args, "only_shortened", False)

    def url_key(item):
        url = item[1][url_pos].strip()

        if not url:
            return

        if only_shortened and not is_shortened_url(url):
            return

        # Url templating
        if cli_args.url_template:
            return cli_args.url_template.format(value=url)

        return url

    def request_args(domain, url, item):
        cookie = None

        # Cookie
        if get_cookie:
            cookie = get_cookie(url)

        # Headers
        headers = None

        if global_headers:
            headers = global_headers

        return {"method": http_method, "cookie": cookie, "headers": headers}

    # Worker callback internals
    filename_builder = None
    files_writer = None

    if not resolve:
        try:
            filename_builder = FilenameBuilder(
                folder_strategy=cli_args.folder_strategy,
                template=cli_args.filename_template,
            )
        except TypeError:
            die(
                [
                    'Invalid "%s" --folder-strategy!' % cli_args.folder_strategy,
                    "Check the list at the end of the command help:",
                    "  $ minet fetch -h",
                ]
            )

        files_writer = ThreadSafeFilesWriter(cli_args.output_dir)

    def worker_callback(result):
        # NOTE: at this point the callback is only fired on success
        row = result.item[1]
        response = result.response
        meta = result.meta

        if cli_args.keep_failed_contents and response.status != 200:
            return

        # First we need to build a filename
        filename_cell = row[filename_pos] if filename_pos is not None else None

        formatter_kwargs = {}

        if cli_args.filename_template and "line" in cli_args.filename_template:
            formatter_kwargs["line"] = enricher.wrap(row)

        try:
            filename = filename_builder(
                result.resolved,
                filename=filename_cell,
                ext=meta.get("ext"),
                formatter_kwargs=formatter_kwargs,
                compressed=cli_args.compress,
            )
        except FilenameFormattingError as e:
            result.error = e
            return

        meta["filename"] = filename

        # Decoding the response data?
        is_text = meta.get("is_text", False)
        original_encoding = meta.get("encoding", "utf-8")

        data = response.data
        binary = True

        if is_text and (cli_args.standardize_encoding or cli_args.contents_in_report):
            data = data.decode(original_encoding, errors="replace")
            binary = False

            if cli_args.contents_in_report:
                meta["decoded_contents"] = data

        # Writing the file?
        # TODO: specify what should happen when contents are empty (e.g. POST queries)
        if data and not cli_args.contents_in_report:
            files_writer.write(
                filename, data, binary=binary, compress=cli_args.compress
            )

    def write_fetch_output(
        index,
        row,
        resolved=None,
        status=None,
        error=None,
        filename=None,
        encoding=None,
        mimetype=None,
        data=None,
    ):

        addendum = [
            resolved or "",
            status or "",
            error or "",
            filename or "",
            mimetype or "",
            encoding or "",
        ]

        if cli_args.contents_in_report:
            addendum.append(data or "")

        enricher.writerow(index, row, addendum)

    def write_resolve_output(
        index, row, resolved=None, status=None, error=None, redirects=None, chain=None
    ):
        addendum = [
            resolved or "",
            status or "",
            error or "",
            redirects or "",
            chain or "",
        ]

        enricher.writerow(index, row, addendum)

    errors = 0
    status_codes = Counter()

    common_kwargs = {
        "key": url_key,
        "insecure": cli_args.insecure,
        "threads": cli_args.threads,
        "throttle": cli_args.throttle,
        "domain_parallelism": cli_args.domain_parallelism,
        "max_redirects": cli_args.max_redirects,
        "wait": False,
        "daemonic": True,
    }

    if cli_args.timeout is not None:
        common_kwargs["timeout"] = cli_args.timeout

    # Normal fetch
    if not resolve:

        multithreaded_iterator = multithreaded_fetch(
            enricher,
            request_args=request_args,
            callback=worker_callback,
            **common_kwargs
        )

        for result in multithreaded_iterator:
            index, row = result.item

            if not result.url:

                write_fetch_output(index, row)

                loading_bar.update()
                continue

            # Updating stats
            update_loading_bar(result)

            # No error
            if result.error is None:
                meta = result.meta

                # Final url target
                resolved_url = result.resolved

                if resolved_url == result.url:
                    resolved_url = None

                # Reporting in output
                write_fetch_output(
                    index,
                    row,
                    resolved=resolved_url,
                    status=result.response.status,
                    filename=meta.get("filename"),
                    encoding=meta.get("encoding"),
                    mimetype=meta.get("mimetype"),
                    data=meta.get("decoded_contents"),
                )

            # Handling potential errors
            else:
                error_code = report_error(result.error)

                resolved = None

                if isinstance(result.error, InvalidURLError):
                    resolved = result.error.url

                if isinstance(result.error, FilenameFormattingError):
                    loading_bar.print(report_filename_formatting_error(result.error))

                write_fetch_output(index, row, error=error_code, resolved=resolved)

    # Resolve
    else:

        multithreaded_iterator = multithreaded_resolve(
            enricher,
            resolve_args=request_args,
            follow_meta_refresh=cli_args.follow_meta_refresh,
            follow_js_relocation=cli_args.follow_js_relocation,
            infer_redirection=cli_args.infer_redirection,
            canonicalize=cli_args.canonicalize,
            **common_kwargs
        )

        for result in multithreaded_iterator:
            index, row = result.item

            if not result.url:

                write_resolve_output(index, row)

                loading_bar.update()
                continue

            # Updating stats
            update_loading_bar(result)

            # No error
            if result.error is None:

                # Reporting in output
                last = result.stack[-1]

                write_resolve_output(
                    index,
                    row,
                    resolved=last.url,
                    status=last.status,
                    redirects=len(result.stack) - 1,
                    chain="|".join(step.type for step in result.stack),
                )

            # Handling potential errors
            else:
                error_code = report_error(result.error)

                write_resolve_output(
                    index,
                    row,
                    error=error_code,
                    redirects=(len(result.stack) - 1) if result.stack else None,
                    chain="|".join(step.type for step in result.stack)
                    if result.stack
                    else None,
                )



# =============================================================================
# Minet BuzzSumo Domain Summary CLI Action
# =============================================================================
#
# Logic of the `bz domain-summary` action.
#
import casanova

from minet.cli.utils import LoadingBar
from minet.buzzsumo import BuzzSumoAPIClient
from minet.buzzsumo.exceptions import (
    BuzzSumoInvalidQueryError,
    BuzzSumoInvalidTokenError,
)

SUMMARY_HEADERS = ["total_results", "total_pages"]


def buzzsumo_domain_summary_action(cli_args):

    client = BuzzSumoAPIClient(cli_args.token)

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        add=SUMMARY_HEADERS,
    )

    loading_bar = LoadingBar(
        desc="Retrieving domain summary", unit="domain", total=enricher.total
    )

    for row, domain_name in enricher.cells(cli_args.column, with_rows=True):

        try:
            data = client.domain_summary(
                domain_name, cli_args.begin_date, cli_args.end_date
            )
        except BuzzSumoInvalidTokenError:
            loading_bar.die("Your BuzzSumo token is invalid!")
        except BuzzSumoInvalidQueryError as e:
            loading_bar.die(
                "Invalid query: %s" % e.url + "\nMessage from the API: %s" % e
            )

        enricher.writerow(row, [data["total_results"], data["total_pages"]])
        loading_bar.update()



# =============================================================================
# Minet Twitter CLI Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the twitter actions.
#
import casanova
import re
from twitter import TwitterHTTPError

from minet.cli.utils import LoadingBar
from minet.cli.exceptions import InvalidArgumentsError
from minet.twitter import TwitterAPIClient

CHARACTERS = re.compile(r"[A-Za-z_]")
NUMBERS = re.compile(r"[0-9]+")
TWITTER_SCREEN_NAME = re.compile(r"[a-zA-Z0-9_]{1,15}")

ITEMS_PER_PAGE = 1000


def validate_query_boundaries(cli_args):
    if cli_args.start_time is not None and cli_args.end_time is not None:
        if cli_args.end_time < cli_args.start_time:
            raise InvalidArgumentsError("--end-time should be after --start-time!")

    if cli_args.since_id and cli_args.until_id:
        if cli_args.until_id < cli_args.since_id:
            raise InvalidArgumentsError("--until-id should be greater than --since-id!")


def make_twitter_action(method_name, csv_headers):
    def action(cli_args):
        enricher = casanova.batch_enricher(
            cli_args.file, cli_args.output, keep=cli_args.select, add=csv_headers
        )

        loading_bar = LoadingBar(
            desc="Retrieving ids", unit=method_name[:-1], stats={"users": 0}
        )

        client = TwitterAPIClient(
            cli_args.access_token,
            cli_args.access_token_secret,
            cli_args.api_key,
            cli_args.api_secret_key,
            api_version="2" if cli_args.v2 else "1.1",
        )

        resuming_state = None

        if cli_args.resume:
            resuming_state = cli_args.output.pop_state()

        for row, user in enricher.cells(cli_args.column, with_rows=True):
            loading_bar.update_stats(user=user)

            all_ids = []
            next_cursor = -1
            result = None

            if cli_args.v2:
                next_cursor = None

            if resuming_state is not None and resuming_state.last_cursor:
                next_cursor = int(resuming_state.last_cursor)

            if cli_args.v2:
                if is_not_user_id(user):
                    loading_bar.die(
                        "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. With --api-v2, you can only use user ids to retrieve followers."
                    )

                client_kwargs = {"max_results": ITEMS_PER_PAGE}

            elif cli_args.ids:
                if is_not_user_id(user):
                    loading_bar.die(
                        "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command."
                    )

                client_kwargs = {"user_id": user}

            else:
                if is_probably_not_user_screen_name(user):
                    loading_bar.die(
                        "The column given as argument probably doesn't contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command."
                    )
                    # force flag to add

                client_kwargs = {"screen_name": user}

            while True:

                skip_in_output = None

                if resuming_state:
                    skip_in_output = resuming_state.values_to_skip
                    resuming_state = None

                if not cli_args.v2:
                    client_kwargs["cursor"] = next_cursor

                    try:
                        result = client.call([method_name, "ids"], **client_kwargs)
                    except TwitterHTTPError as e:

                        # The user does not exist
                        loading_bar.inc("users_not_found")
                        break

                    if result is not None:
                        all_ids = result.get("ids", [])
                        next_cursor = result.get("next_cursor", 0)

                        loading_bar.update(len(all_ids))

                        batch = []

                        for user_id in all_ids:
                            if skip_in_output and user_id in skip_in_output:
                                continue

                            batch.append([user_id])

                    else:
                        break

                else:
                    if method_name == "friends":
                        method_name_v2 = "following"
                    else:
                        method_name_v2 = method_name

                    try:
                        result = client.call(
                            route=["users", user, method_name_v2], **client_kwargs
                        )
                    except TwitterHTTPError as e:

                        # The user does not exist
                        loading_bar.inc("users_not_found")
                        break

                    if result is not None and "data" in result:
                        batch = []

                        for follower_metadata in result["data"]:
                            user_id = follower_metadata["id"]

                            if skip_in_output and user_id in skip_in_output:
                                continue
                            batch.append([user_id])

                        loading_bar.update(len(result["data"]))

                        if "next_token" in result["meta"]:
                            next_cursor = result["meta"]["next_token"]
                            client_kwargs["pagination_token"] = next_cursor
                        else:
                            next_cursor = None

                    else:
                        break

                enricher.writebatch(row, batch, next_cursor or None)

                if next_cursor is None or next_cursor == 0:
                    break

            loading_bar.inc("users")

    return action


def is_not_user_id(item):
    return bool(re.match(CHARACTERS, item))


def is_probably_not_user_screen_name(item):
    matches = TWITTER_SCREEN_NAME.fullmatch(item)
    if matches:
        return bool(NUMBERS.fullmatch(item))
    return True



from minet.__version__ import __version__

from minet.crawl import Crawler, CrawlJob, Spider, BeautifulSoupSpider, DefinitionSpider
from minet.fetch import multithreaded_fetch, multithreaded_resolve
from minet.scrape import Scraper



from setuptools import setup, find_packages

with open("./README.md", "r") as f:
    long_description = f.read()

# Version
# Info: https://packaging.python.org/guides/single-sourcing-package-version/
# Example: https://github.com/pypa/warehouse/blob/64ca42e42d5613c8339b3ec5e1cb7765c6b23083/warehouse/__about__.py
meta_package = {}
with open("./minet/__version__.py") as f:
    exec(f.read(), meta_package)


setup(
    name="minet",
    version=meta_package["__version__"],
    description="A webmining CLI tool & library for python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/medialab/minet",
    license="MIT",
    author="Jules Farjas, Guillaume Plique, Pauline Breteau",
    keywords="webmining",
    python_requires=">=3.6",
    packages=find_packages(exclude=["ftest", "scripts", "test"]),
    install_requires=[
        "beautifulsoup4>=4.7.1",
        "browser-cookie3==0.16.1",
        "casanova>=0.18.0,<0.19",
        "cchardet>=2.1.7",
        "colorama>=0.4.0",
        "dateparser>=1.1.1",
        "ebbe>=1.9.0,<2",
        "json5>=0.8.5",
        "keyring<19.3",
        "lxml>=4.3.0",
        "ndjson>=0.3.1",
        "persist-queue>=0.7.0",
        "pyyaml",
        "quenouille>=1.4.3,<2",
        "soupsieve>=2.1",
        "tenacity>=7.0.0",
        "termcolor>=1.0.0",
        "tqdm>=4.60.0",
        "trafilatura>=1.2.0,<1.3",
        "twitwi>=0.14.1,<0.15",
        "ural>=0.34.0,<0.35",
        "urllib3>=1.26.9,<2",
    ],
    entry_points={"console_scripts": ["minet=minet.cli.__main__:main"]},
    zip_safe=True,
)



# =============================================================================
# Minet Hyphe Dump CLI Action
# =============================================================================
#
# Logic of the `hyphe dump` action.
#
import os
import csv
import base64
import zlib
import gzip
from os.path import join, dirname

from minet.utils import md5
from minet.cli.utils import die, LoadingBar
from minet.hyphe import HypheAPIClient
from minet.hyphe.formatters import format_webentity_for_csv, format_page_for_csv
from minet.hyphe.constants import WEBENTITY_CSV_HEADERS, PAGE_CSV_HEADERS
from minet.hyphe.exceptions import HypheCorpusAuthenticationError

ADDITIONAL_PAGE_HEADERS = ["filename"]


def format_page_filename(webentity, page):
    h = md5(page["url"])

    # TODO: could be something other than html?
    return "%s/%s/%s.html.gz" % (webentity["id"], h[:2], h)


def hyphe_dump_action(cli_args):

    # Paths
    output_dir = "hyphe_corpus_%s" % cli_args.corpus

    if cli_args.output_dir is not None:
        output_dir = cli_args.output_dir

    os.makedirs(output_dir, exist_ok=True)

    webentities_output_path = join(output_dir, "webentities.csv")
    pages_output_path = join(output_dir, "pages.csv")

    if cli_args.body:
        body_output_dir = join(output_dir, "content")
        os.makedirs(body_output_dir, exist_ok=True)

    client = HypheAPIClient(cli_args.url)
    corpus = client.corpus(cli_args.corpus, password=cli_args.password)

    try:
        corpus.ensure_is_started()
    except HypheCorpusAuthenticationError:
        die(
            [
                'Wrong password for the "%s" corpus!' % cli_args.corpus,
                "Don't forget to provide a password for this corpus using --password",
            ]
        )

    # Then we gather some handy statistics
    counts = corpus.count(statuses=cli_args.statuses)

    # Then we fetch webentities
    webentities_file = open(webentities_output_path, "w", encoding="utf-8")
    webentities_writer = csv.writer(webentities_file)
    webentities_writer.writerow(WEBENTITY_CSV_HEADERS)

    loading_bar = LoadingBar(
        desc="Paginating web entities",
        unit="webentity",
        unit_plural="webentities",
        total=counts["webentities"],
    )

    webentities = {}

    for webentity in corpus.webentities(statuses=cli_args.statuses):
        loading_bar.update()
        webentities[webentity["id"]] = webentity
        webentities_writer.writerow(format_webentity_for_csv(webentity))

    webentities_file.close()
    loading_bar.close()

    # Finally we paginate pages
    pages_file = open(pages_output_path, "w", encoding="utf-8")
    pages_writer = csv.writer(pages_file)
    pages_writer.writerow(
        PAGE_CSV_HEADERS + (ADDITIONAL_PAGE_HEADERS if cli_args.body else [])
    )

    loading_bar = LoadingBar(desc="Fetching pages", unit="page", total=counts["pages"])

    for webentity in webentities.values():
        for page in corpus.webentity_pages(webentity["id"], include_body=cli_args.body):
            loading_bar.update()

            filename = None

            if cli_args.body and "body" in page:
                filename = format_page_filename(webentity, page)
                filepath = join(body_output_dir, filename)
                os.makedirs(dirname(filepath), exist_ok=True)

                with open(filter, "wb") as f:
                    binary = base64.b64decode(page["body"])
                    binary = zlib.decompress(binary)
                    binary = gzip.compress(binary)  # TODO: use gzip.open rather

                    f.write(binary)

            pages_writer.writerow(
                format_page_for_csv(webentity, page, filename=filename)
            )



# =============================================================================
# Minet Hyphe CLI Action
# =============================================================================
#
# Logic of the `hyphe` action.
#


def hyphe_action(cli_args):

    if cli_args.hyphe_action == "declare":
        from minet.cli.hyphe.declare import hyphe_declare_action

        hyphe_declare_action(cli_args)

    elif cli_args.hyphe_action == "destroy":
        from minet.cli.hyphe.destroy import hyphe_destroy_action

        hyphe_destroy_action(cli_args)

    elif cli_args.hyphe_action == "dump":
        from minet.cli.hyphe.dump import hyphe_dump_action

        hyphe_dump_action(cli_args)

    elif cli_args.hyphe_action == "reset":
        from minet.cli.hyphe.reset import hyphe_reset_action

        hyphe_reset_action(cli_args)

    elif cli_args.hyphe_action == "tag":
        from minet.cli.hyphe.tag import hyphe_tag_action

        hyphe_tag_action(cli_args)



# =============================================================================
# Minet CLI Reporters
# =============================================================================
#
# Various reporters whose goal is to convert errors etc. into human-actionable
# labels in CSV format, for instance.
#
from io import StringIO
from termcolor import colored
from functools import partial
from urllib3.exceptions import (
    ConnectTimeoutError,
    MaxRetryError,
    ReadTimeoutError,
    ResponseError,
    SSLError,
    NewConnectionError,
    ProtocolError,
    DecodeError,
)

from minet.exceptions import (
    UnknownEncodingError,
    InvalidURLError,
    MaxRedirectsError,
    InfiniteRedirectsError,
    SelfRedirectError,
    InvalidRedirectError,
    TrafilaturaError,
    FilenameFormattingError,
)
from minet.scrape.constants import BURROWING_KEYS, LEAF_KEYS
from minet.scrape.exceptions import (
    ScraperEvalSyntaxError,
    ScraperValidationConflictError,
    InvalidCSSSelectorError,
    ScraperValidationIrrelevantPluralModifierError,
    ScraperValidationMixedConcernError,
    ScraperValidationInvalidPluralModifierError,
    ScraperValidationInvalidExtractorError,
    ScraperValidationUnknownKeyError,
    ScraperEvalError,
)


def max_retry_error_reporter(error):
    if isinstance(error, ConnectTimeoutError):
        return "connect-timeout"

    if isinstance(error, ReadTimeoutError):
        return "read-timeout"

    if isinstance(error.reason, ResponseError) and "redirect" in repr(error.reason):
        return "too-many-redirects"

    return "max-retries-exceeded"


def new_connection_error_reporter(error):
    msg = repr(error).lower()

    if "name or service not known" in msg or "errno 8" in msg:
        return "unknown-host"

    if "connection refused" in msg:
        return "connection-refused"

    return msg


def protocol_error_reporter(error):
    msg = repr(error).lower()

    if "connection aborted" in msg:
        return "connection-aborted"

    if "connection refused" in msg:
        return "connection-refused"

    return "connection-error"


def decode_error_reporter(error):
    msg = repr(error)

    if "gzip" in msg:
        return "invalid-gzip"

    return "invalid-encoding"


ERROR_REPORTERS = {
    UnicodeDecodeError: "wrong-encoding",
    UnknownEncodingError: "unknown-encoding",
    FileNotFoundError: "file-not-found",
    MaxRetryError: max_retry_error_reporter,
    InvalidURLError: "invalid-url",
    SSLError: "ssl",
    NewConnectionError: new_connection_error_reporter,
    ProtocolError: protocol_error_reporter,
    ConnectTimeoutError: "connect-timeout",
    ReadTimeoutError: "read-timeout",
    MaxRedirectsError: "max-redirects",
    InfiniteRedirectsError: "infinite-redirects",
    SelfRedirectError: "self-redirect",
    InvalidRedirectError: "invalid-redirect",
    DecodeError: decode_error_reporter,
    TrafilaturaError: "trafilatura-error",
    FilenameFormattingError: "filename-formatting-error",
}


def report_error(error):
    reporter = ERROR_REPORTERS.get(type(error), repr)

    return reporter(error) if callable(reporter) else reporter


def and_join(strings):
    strings = [colored(s, "green") for s in strings]

    if len(strings) < 2:
        return strings[0]

    return (", ".join(strings[:-1])) + " and " + strings[-1]


def report_scraper_validation_errors(errors):
    output = StringIO()

    p = partial(print, file=output)

    red_alert = colored("Error", "red")

    for n, error in enumerate(errors, 1):
        path = "." + (".".join(error.path))

        p(
            "> {error} n°{n} at path {path}{root}".format(
                error=red_alert,
                n=n,
                path=colored(path, "blue"),
                root=(" (root)" if not error.path else ""),
            )
        )

        if isinstance(error, ScraperValidationConflictError):
            p(
                "  the {keys} keys are conflicting and should not be found at the same level!".format(
                    keys=and_join(error.keys)
                )
            )

        elif isinstance(error, ScraperValidationIrrelevantPluralModifierError):
            p(
                "  the {modifier} modifier should not be found at a non-plural level (i.e. without iterator)!".format(
                    modifier=colored(error.modifier, "green")
                )
            )

        elif isinstance(error, ScraperValidationInvalidPluralModifierError):
            p(
                "  the {modifier} modifier cannot be a boolean without {fields} and cannot be a key/path with {fields}!".format(
                    modifier=colored(error.modifier, "green"),
                    fields=colored("fields", "green"),
                )
            )

        elif isinstance(error, ScraperValidationInvalidExtractorError):
            p(
                "  unknown {extractor} extractor!".format(
                    extractor=colored(error.extractor, "green")
                )
            )

        elif isinstance(error, ScraperValidationMixedConcernError):
            p(
                "  mixed concerns could not be interpreted (i.e. the {burrowing} keys should not be found alongside the {leaf} ones)!".format(
                    burrowing=and_join(BURROWING_KEYS), leaf=and_join(LEAF_KEYS)
                )
            )

        elif isinstance(error, InvalidCSSSelectorError):
            p(
                "  invalid CSS selector {css}".format(
                    css=colored(error.expression, "cyan")
                )
            )

        elif isinstance(error, ScraperEvalSyntaxError):
            p("  invalid python code was found:")

            for line in error.expression.split("\n"):
                p(colored("    | {line}".format(line=line), "cyan"))

        elif isinstance(error, ScraperValidationUnknownKeyError):
            p("  unknown {key} key!".format(key=colored(error.key, "cyan")))

        p()

    return output.getvalue()


def report_scraper_evaluation_error(error):
    output = StringIO()

    p = partial(print, file=output)

    red_alert = colored("Scraper error", "red")

    path = "." + (".".join(error.path))

    p(
        "> {error} at path {path}{root}".format(
            error=red_alert,
            path=colored(path, "blue"),
            root=(" (root)" if not error.path else ""),
        )
    )

    if isinstance(error, ScraperEvalError):
        msg = str(error.reason)
        p(
            "  evaluated code raised {error}{msg}!".format(
                error=colored(error.reason.__class__.__name__, "green"),
                msg=(" with message: " + colored(msg, "magenta")) if msg else "",
            )
        )

        for line in error.expression.split("\n"):
            p(colored("    | {line}".format(line=line), "cyan"))

    p()

    return output.getvalue()


def report_filename_formatting_error(error):
    return "> error when formatting filename using: {template}".format(
        template=colored(error.template, "cyan")
    )



# =============================================================================
# Minet Facebook Posts CLI Action
# =============================================================================
#
# Logic of the `fb posts` action.
#
import casanova

from minet.constants import COOKIE_BROWSERS
from minet.cli.utils import die, LoadingBar
from minet.facebook import FacebookMobileScraper
from minet.facebook.constants import FACEBOOK_POST_CSV_HEADERS
from minet.facebook.exceptions import (
    FacebookInvalidCookieError,
    FacebookInvalidTargetError,
)


def facebook_posts_action(cli_args):
    try:
        scraper = FacebookMobileScraper(cli_args.cookie, throttle=cli_args.throttle)
    except FacebookInvalidCookieError:
        if cli_args.cookie in COOKIE_BROWSERS:
            die(['Could not extract relevant cookie from "%s".' % cli_args.cookie])

        die(
            [
                "Relevant cookie not found.",
                "A Facebook authentication cookie is necessary to be able to scrape Facebook groups.",
                "Use the --cookie flag to choose a browser from which to extract the cookie or give your cookie directly.",
            ]
        )

    # Enricher
    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=FACEBOOK_POST_CSV_HEADERS,
    )

    # Loading bar
    loading_bar = LoadingBar(desc="Scraping posts", unit="post")

    translated_langs = set()

    for i, (row, url) in enumerate(enricher.cells(cli_args.column, with_rows=True), 1):
        loading_bar.inc("groups")

        try:
            posts = scraper.posts(url)
        except FacebookInvalidTargetError:
            loading_bar.print(
                "Given url (line %i) is probably not a Facebook group: %s" % (i, url)
            )
            continue

        for post in posts:
            if post.translated_text and post.translated_from not in translated_langs:
                translated_langs.add(post.translated_from)
                lines = [
                    "Found text translated from %s!" % post.translated_from,
                    "Since it means original text may not be entirely retrieved you might want",
                    'to edit your Facebook language settings to add "%s" to'
                    % post.translated_from,
                    'the "Languages you don\'t want to be offered translations for" list here:',
                    "https://www.facebook.com/settings/?tab=language",
                ]

                for line in lines:
                    loading_bar.print(line)

                loading_bar.print()

            loading_bar.update()
            enricher.writerow(row, post.as_csv_row())



# =============================================================================
# Minet Youtube Channel CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and retrieving metadata about
# the given Youtube channels using Google's APIs.
#
import casanova

from minet.cli.utils import LoadingBar
from minet.youtube import YouTubeAPIClient
from minet.youtube.constants import YOUTUBE_PLAYLIST_VIDEO_SNIPPET_CSV_HEADERS


def channel_videos_action(cli_args):
    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        add=YOUTUBE_PLAYLIST_VIDEO_SNIPPET_CSV_HEADERS,
        keep=cli_args.select,
    )

    loading_bar = LoadingBar(desc="Retrieving videos", unit="video")

    def before_sleep_until_midnight(seconds):
        loading_bar.print(
            "API limits reached. Will now wait until midnight Pacific time!"
        )

    client = YouTubeAPIClient(
        cli_args.key, before_sleep_until_midnight=before_sleep_until_midnight
    )

    for row, channel_id in enricher.cells(cli_args.column, with_rows=True):
        loading_bar.print('Retrieving videos for "%s"' % channel_id)
        for video in client.channel_videos(channel_id):
            loading_bar.update()
            enricher.writerow(row, video.as_csv_row())

        loading_bar.inc("channels")



# =============================================================================
# Minet Youtube Scrapers
# =============================================================================
#
# Collection of YouTube-related scrapers.
#
import re
import json
from bs4 import BeautifulSoup
from html import unescape
from urllib.parse import unquote
from collections import namedtuple

from minet.web import request, create_pool
from minet.youtube.utils import ensure_video_id
from minet.youtube.exceptions import YouTubeInvalidVideoTargetError

CAPTION_TRACKS_RE = re.compile(r'({"captionTracks":.*isTranslatable":(true|false)}])')
TIMEDTEXT_RE = re.compile(rb'timedtext?[^"]+')

YOUTUBE_SCRAPER_POOL = create_pool()

YouTubeCaptionTrack = namedtuple("YouTubeCaptionTrack", ["lang", "url", "generated"])


def get_caption_tracks(video_id):

    # First we try to retrieve it from video info
    url = "https://www.youtube.com/get_video_info?video_id=%s" % video_id

    err, response = request(url)

    if err:
        raise err

    data = unquote(response.data.decode("utf-8"))

    m = CAPTION_TRACKS_RE.search(data)

    if m is not None:
        data = json.loads(m.group(0) + "}")["captionTracks"]

        return [
            YouTubeCaptionTrack(
                item["languageCode"], item["baseUrl"], item.get("kind") == "asr"
            )
            for item in data
        ]

    # Then we try to scrape it directly from the video page
    # url = 'https://www.youtube.com/watch?v=%s' % video_id

    # err, response = request(url)

    # if err:
    #     raise err

    # timedtexts = TIMEDTEXT_RE.findall(response.data)

    return []


def select_caption_track(tracks, langs=None, strict=True):
    def key(t):
        try:
            return (langs.index(t.lang), int(t.generated), t.lang)
        except ValueError:
            return (len(langs), int(t.generated), t.lang)

    best = min(tracks, key=key)

    if strict and best.lang not in langs:
        return None

    return best


def get_video_captions(video_target, langs):
    if not isinstance(langs, list):
        raise TypeError

    video_id = ensure_video_id(video_target)

    if video_id is None:
        raise YouTubeInvalidVideoTargetError

    tracks = get_caption_tracks(video_id)

    if not tracks:
        return

    best_track = select_caption_track(tracks, langs=langs)

    if best_track is None:
        return

    err, response = request(best_track.url, pool=YOUTUBE_SCRAPER_POOL)

    if err:
        raise err

    soup = BeautifulSoup(response.data.decode("utf-8"), "lxml")

    captions = []

    for item in soup.select("text"):
        captions.append(
            (item.get("start"), item.get("dur"), unescape(item.get_text().strip()))
        )

    return best_track, captions


def scrape_channel_id(channel_url):
    err, response = request(channel_url, pool=YOUTUBE_SCRAPER_POOL)

    if err:
        raise err

    soup = BeautifulSoup(response.data.decode("utf-8"), "lxml")
    channel_id = soup.find("meta", {"itemprop": "channelId"}).get("content")

    return channel_id
